/*
 * Copyright (C) 2004 Ivo Danihelka (ivo@danihelka.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */
#include "InputAgent.h"

#include "KeyBinder.h"
#include "InputHandler.h"

#include "MessagerAgent.h"
#include "SimpleMsg.h"
#include "UnknownMsgException.h"
#include "Name.h"
#include "MouseStroke.h"

#include "SDL.h"

//-----------------------------------------------------------------
/**
 * Enable SDL_UNICODE.
 * Enable key repeat.
 * Set console handler to ScriptAgent.
 *
 * NOTE: every SDL_InitSubSystem will disable UNICODE
 * hence InputAgent init must be after VideoAgent and SoundAgent.
 * NOTE: KeyConsole() use Path which asks OptionAgent
 */
    void
InputAgent::own_init()
{
    m_keyBinder = new KeyBinder();
    m_handler = NULL;
    m_keys = SDL_GetKeyState(NULL);

    SDL_EnableUNICODE(1);
}

// ZX
# if defined( WIZ_MODE ) 

#define GP2X_UP              (0)
#define GP2X_DOWN            (4)
#define GP2X_LEFT            (2)
#define GP2X_RIGHT           (6)
#define GP2X_UPLEFT          (1)
#define GP2X_UPRIGHT         (7)
#define GP2X_DOWNLEFT        (3)
#define GP2X_DOWNRIGHT       (5)

#define GP2X_A               (12)
#define GP2X_B               (13)
#define GP2X_X               (14)
#define GP2X_Y               (15)
#define GP2X_L               (10)
#define GP2X_R               (11)
#define GP2X_FIRE            (18)
#define GP2X_MENU            (8)        
#define GP2X_SELECT          (9)
#define GP2X_VOLUP           (16)
#define GP2X_VOLDOWN         (17)

static void
loc_ConvertButtonToKeyEvent( SDL_Event& event )
{
  int event_type = event.type;

# if 0
fprintf(stdout, "loc_ConvertButtonToKeyEvent %d\n", event_type);
# endif

  if ((event_type != SDL_JOYBUTTONDOWN ) &&
      (event_type != SDL_JOYBUTTONUP   )) return;

  int button = event.jbutton.button;

# if 0
fprintf(stdout, "button: %d\n", button );
# endif

  SDL_KeyboardEvent* kbd_event = (SDL_KeyboardEvent *)&event;

  if (event_type == SDL_JOYBUTTONDOWN) {
    kbd_event->type = SDL_KEYDOWN;
    kbd_event->state = SDL_PRESSED;
  } else {
    kbd_event->type = SDL_KEYUP;
    kbd_event->state = SDL_RELEASED;
  }
  kbd_event->which = 0;

  int key_sym = 0;

  switch ( button ) {
      case GP2X_UP        : key_sym = SDLK_UP;
      break;
      case GP2X_DOWN      : key_sym = SDLK_DOWN;
      break;
      case GP2X_LEFT      : key_sym = SDLK_LEFT;
      break;
      case GP2X_RIGHT     : key_sym = SDLK_RIGHT;
      break;
      case GP2X_A         : key_sym = SDLK_RETURN;
      break;
      case GP2X_B         : key_sym = SDLK_F5;
      break;
      case GP2X_X         : key_sym = SDLK_SPACE;
      break;
      case GP2X_Y         : key_sym = SDLK_BACKSPACE;
      break;
      case GP2X_L         : key_sym = SDLK_F2;
      break;
      case GP2X_R         : key_sym = SDLK_F3;
      break;
      case GP2X_MENU      : key_sym = SDLK_F10;
      break;
      case GP2X_SELECT    : key_sym = SDLK_ESCAPE;
      break;
# if 0
      case GP2X_VOLUP     : key_sym = GP2X_CTRL_VOLUP;
      break;
      case GP2X_VOLDOWN   : key_sym = GP2X_CTRL_VOLDOWN;
      break;
# endif
  }
  kbd_event->keysym.scancode = button; // don't care
  kbd_event->keysym.sym      = SDLKey(key_sym);
  kbd_event->keysym.mod      = SDLMod(0);
  kbd_event->keysym.unicode  = key_sym;

# if 0
fprintf(stdout, "keysym: %d\n", key_sym );
# endif
}
# endif

//-----------------------------------------------------------------
    void
InputAgent::own_update()
{
# if defined( WIZ_MODE )
static SDL_Event loc_last_joy_event;
static long      loc_last_joy_timestamp;
# endif

    SDL_Event event;

    while (SDL_PollEvent(&event)) {
# if defined(WIZ_MODE)
        switch (event.type) {
            case SDL_JOYBUTTONDOWN :
               loc_last_joy_event = event;
               loc_last_joy_timestamp = SDL_GetTicks() * 1000;
               loc_ConvertButtonToKeyEvent( event );
            break;
            case SDL_JOYBUTTONUP :
               loc_last_joy_event = event;
               loc_last_joy_timestamp = SDL_GetTicks() * 1000;
               loc_ConvertButtonToKeyEvent( event );
            break;
         }
# endif
        switch (event.type) {
            case SDL_QUIT:
                {
                    BaseMsg *msg = new SimpleMsg(Name::APP_NAME, "quit");
                    MessagerAgent::agent()->forwardNewMsg(msg);
                    break;
                }
            case SDL_KEYDOWN:
                m_keyBinder->keyDown(event.key.keysym);
                if (m_handler) {
                    m_handler->keyEvent(KeyStroke(event.key.keysym));
                }
                break;
            case SDL_KEYUP:
                if (m_handler) {
                    m_handler->keyUp(KeyStroke(event.key.keysym));
                }
                break;
            case SDL_MOUSEBUTTONDOWN:
                if (m_handler) {
                    m_handler->mouseEvent(MouseStroke(event.button));
                }
                break;
            default:
                break;
        }
    }
# if defined(WIZ_MODE) 
# define GP2X_MIN_TIME_REPEAT  150000
    if (loc_last_joy_event.type == SDL_JOYBUTTONDOWN) {
      long timestamp = SDL_GetTicks() * 1000;
      if ((timestamp - loc_last_joy_timestamp) > GP2X_MIN_TIME_REPEAT) {
         loc_last_joy_timestamp = timestamp;
         event = loc_last_joy_event;
         loc_ConvertButtonToKeyEvent( event );
         if (event.type == SDL_KEYDOWN) {
           m_keyBinder->keyDown(event.key.keysym);
           if (m_handler) {
               m_handler->keyEvent(KeyStroke(event.key.keysym));
           }
         }
      }
    }
# endif
    if (m_handler) {
        Uint8 buttons;
        V2 mouseLoc = getMouseState(&buttons);
        m_handler->mouseState(mouseLoc, buttons);
    }
}
//-----------------------------------------------------------------
/**
 * Delete console.
 */
    void
InputAgent::own_shutdown()
{
    delete m_keyBinder;
}
//-----------------------------------------------------------------
void
InputAgent::installHandler(InputHandler *handler)
{
    if (m_handler) {
        m_handler->takePressed(NULL);
        m_handler->mouseState(V2(-1, -1), 0);
    }
    m_handler = handler;
    if (m_handler) {
        m_handler->takePressed(m_keys);
        Uint8 buttons;
        V2 mouseLoc = getMouseState(&buttons);
        m_handler->mouseState(mouseLoc, buttons);
    }
}
//-----------------------------------------------------------------
/**
 * Return mouse location.
 * @param out_buttons place where to store state of buttons
 * @return (mouseX, mouseY)
 */
    V2
InputAgent::getMouseState(Uint8 *out_buttons)
{
    int x;
    int y;
    Uint8 pressed = SDL_GetMouseState(&x, &y);
    if (out_buttons) {
        *out_buttons = pressed;
    }
#if defined( WIZ_MODE )
    int re_x = int(float(x) * MouseStroke::getScaleX());
    int re_y = int(float(y) * MouseStroke::getScaleY());
    return V2(re_x, re_y);
# else
    return V2(x, y);
# endif
}


